/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2017,2018 Marc Alexander Lehmann / the Deliantra team
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001-2003 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992 Frank Tore Johansen
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <unistd.h>

#include "global.h"
#include "path.h"

//+GPL

sint8 maptile::outdoor_darkness;

/* This rolls up wall, blocks_magic, blocks_view, etc, all into
 * one function that just returns a P_.. value (see map.h)
 * it will also do map translation for tiled maps, returning
 * new values into newmap, nx, and ny.  Any and all of those
 * values can be null, in which case if a new map is needed (returned
 * by a P_NEW_MAP value, another call to get_map_from_coord
 * is needed.  The case of not passing values is if we're just
 * checking for the existence of something on those spaces, but
 * don't expect to insert/remove anything from those spaces.
 */
int
get_map_flags (maptile *oldmap, maptile **newmap, sint16 x, sint16 y, sint16 *nx, sint16 *ny)
{
  sint16 newx = x;
  sint16 newy = y;

  maptile *mp = get_map_from_coord (oldmap, &newx, &newy);

  if (!mp)
    return P_OUT_OF_MAP;

  if (newmap) *newmap = mp;
  if (nx)     *nx     = newx;
  if (ny)     *ny     = newy;

  return mp->at (newx, newy).flags () | (mp != oldmap ? P_NEW_MAP : 0);
}

/*
 * Returns true if the given coordinate is blocked except by the
 * object passed is not blocking.  This is used with
 * multipart monsters - if we want to see if a 2x2 monster
 * can move 1 space to the left, we don't want its own area
 * to block it from moving there.
 * Returns TRUE if the space is blocked by something other than the
 * monster.
 * m, x, y are the target map/coordinates - needed for map tiling.
 * the coordinates & map passed in should have been updated for tiling
 * by the caller.
 */
int
blocked_link (object *ob, maptile *m, int sx, int sy)
{
  /* Make sure the coordinates are valid - they should be, as caller should
   * have already checked this.
   */
  if (OUT_OF_REAL_MAP (m, sx, sy))
    {
      LOG (llevError | logBacktrace, "blocked_link: Passed map, x, y coordinates outside of map\n");
      return 1;
    }

  mapspace &ms = m->at (sx, sy);

  int mflags  = ms.flags ();
  int blocked = ms.move_block;

  /* If space is currently not blocked by anything, no need to
   * go further.  Not true for players - all sorts of special
   * things we need to do for players.
   */
  if (ob->type != PLAYER && !(mflags & P_IS_ALIVE) && blocked == 0)
    return 0;

  /* if there isn't anything alive on this space, and this space isn't
   * otherwise blocked, we can return now.  Only if there is a living
   * creature do we need to investigate if it is part of this creature
   * or another.  Likewise, only if something is blocking us do we
   * need to investigate if there is a special circumstance that would
   * let the player through (inventory checkers for example)
   */
  if (!(mflags & P_IS_ALIVE) && !OB_TYPE_MOVE_BLOCK (ob, blocked))
    return 0;

  ob = ob->head_ ();

  /* We basically go through the stack of objects, and if there is
   * some other object that has NO_PASS or FLAG_ALIVE set, return
   * true.  If we get through the entire stack, that must mean
   * ob is blocking it, so return 0.
   */
  for (object *tmp = ms.top; tmp; tmp = tmp->below)
    {
      if (OB_MOVE_BLOCK (ob, tmp))
        {
          if (INVOKE_OBJECT (BLOCKED_MOVE, tmp, ob))
            if (RESULT_INT (0))
              return 1;
            else
              continue;

          if (tmp->type == CHECK_INV)
            {
              bool have = check_inv_recursive (ob, tmp);

              // last_sp set means we block if we don't have.
              if (logical_xor (have, tmp->last_sp))
                return 1;
            }
          else if (tmp->type == T_MATCH)
            {
              // T_MATCH allows "entrance" iff the match is true
              // == blocks if the match fails

              // we could have used an INVOKE_OBJECT, but decided against it, as we
              // assume that T_MATCH is relatively common.
              if (!match (tmp->slaying, ob, tmp, ob))
                return 1;
            }
          else
            return 1; // unconditional block

        } else {
          // space does not block the ob, directly, but
          // anything alive that is not a door still
          // blocks anything

          if (tmp->flag [FLAG_ALIVE]
              && tmp->type != DOOR
              && tmp->head_ () != ob) //TODO: maybe move these check up?
            return 1;
        }
    }

  return 0;
}

/*
 * Returns the blocking object if the given object can't fit in the given
 * spot.  This is meant for multi space objects - for single space objecs,
 * just calling get_map_blocked and checking that against movement type
 * of object. This function goes through all the parts of the multipart
 * object and makes sure they can be inserted.
 *
 * While this doesn't call out of map, the get_map_flags does.
 *
 * This function has been used to deprecate arch_out_of_map -
 * this function also does that check, and since in most cases,
 * a call to one would follow the other, doesn't make a lot of sense to
 * have two seperate functions for this.
 *
 * This returns nonzero if this arch can not go on the space provided,
 * 0 otherwise.  the return value will contain the P_.. value
 * so the caller can know why this object can't go on the map.
 * Note that callers should not expect P_NEW_MAP to be set
 * in return codes - since the object is multispace - if
 * we did return values, what do you return if half the object
 * is one map, half on another.
 *
 * Note this used to be arch_blocked, but with new movement
 * code, we need to have actual object to check its move_type
 * against the move_block values.
 */
bool
object::blocked (maptile *m, int x, int y) const
{
  for (archetype *tmp = arch; tmp; tmp = (archetype *)tmp->more)
    {
      mapxy pos (m, x + tmp->x, y + tmp->y);

      if (!pos.normalise ())
        return 1;

      mapspace &ms = *pos;

      if (ms.flags () & P_IS_ALIVE)
        return 1;

      /* However, often  ob doesn't have any move type
       * (signifying non-moving objects)
       * so the AND operation in OB_TYPE_MOVE_BLOCK doesn't work.
       */
      if (!move_type && ms.move_block != MOVE_ALL)
        continue;

      /* Note it is intentional that we check ob - the movement type of the
       * head of the object should correspond for the entire object.
       */
      if (ms.blocks (move_type))
        return 1;
    }

  return 0;
}

//-GPL

void
maptile::set_object_flag (int flag, int value)
{
  if (!spaces)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *tmp = ms->bot; tmp; tmp = tmp->above)
      tmp->flag [flag] = value;
}

void
maptile::post_load_original ()
{
  if (!spaces)
    return;

  set_object_flag (FLAG_OBJ_ORIGINAL);

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *tmp = ms->bot; tmp; tmp = tmp->above)
      INVOKE_OBJECT (RESET, tmp);
}

void
maptile::post_load ()
{
#if 0
  if (!spaces)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *tmp = ms->bot; tmp; tmp = tmp->above)
      ; // nop
#endif
}

//+GPL

/* link_multipart_objects go through all the objects on the map looking
 * for objects whose arch says they are multipart yet according to the
 * info we have, they only have the head (as would be expected when
 * they are saved).
 */
void
maptile::link_multipart_objects ()
{
  if (!spaces)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    {
      object *op = ms->bot;
      while (op)
        {
          /* already multipart - don't do anything more */
          if (op->head_ () == op && !op->more && op->arch->more)
            {
              op->remove ();
              op->expand_tail ();

              // FIXME: INS_ON_TOP is just a workaround for the pentagram vs.
              // multi-tile monster bug, where INS_ABOVE_FLOOR_ONLY put the monsters
              // below the pentagrams... hopefully INS_ON_TOP doesn't break anything
              insert (op, op->x, op->y, 0, INS_NO_MERGE | INS_ON_TOP | INS_NO_WALK_ON);

              op = ms->bot; // we are mutating the mapspace too much with INS_ON_TOP
                            // so we have to reset the iteration through the mapspace
            }
          else
            op = op->above;
        }
    }
}

//-GPL

/*
 * Loads (ands parses) the objects into a given map from the specified
 * file pointer.
 */
bool
maptile::_load_objects (object_thawer &f)
{
  for (;;)
    {
      coroapi::cede_to_tick (); // cede once in a while

      switch (f.kw)
        {
          case KW_arch:
            if (object *op = object::read (f, this))
              {
                // TODO: why?
                if (op->inv)
                  {
                    op->carrying = 0;
                    op->update_weight ();
                  }

                if (IN_RANGE_EXC (op->x, 0, width) && IN_RANGE_EXC (op->y, 0, height))
                  {
                    // we insert manually because
                    // a) its way faster
                    // b) we remove manually, too, and there are good reasons for that
                    // c) it's correct
                    mapspace &ms = at (op->x, op->y);

                    op->flag [FLAG_REMOVED] = false;

                    op->above = 0;
                    op->below = ms.top;

                    *(ms.top ? &ms.top->above : &ms.bot) = op;

                    ms.top = op;
                    ms.flags_ = 0;
                  }
                else
                  {
                    f.parse_warn (format ("object %s out of range", op->debug_desc ()));
                    op->destroy ();
                  }
              }

            continue;

          case KW_EOF:
            return true;

          default:
            if (!f.parse_error ("map file"))
              return false;
            break;
        }

        f.next ();
    }

  return true;
}

void
maptile::activate ()
{
  if (state != MAP_INACTIVE)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *op = ms->bot; op; op = op->above)
      op->activate_recursive ();

  state = MAP_ACTIVE;

  activate_physics ();
}

void
maptile::deactivate ()
{
  if (state != MAP_ACTIVE)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *op = ms->bot; op; op = op->above)
      op->deactivate_recursive ();

  state = MAP_INACTIVE;
}

bool
maptile::_save_objects (object_freezer &f, int flags)
{
  coroapi::cede_to_tick ();

  if (flags & IO_HEADER)
    _save_header (f);

  if (!spaces)
    return false;

  for (int i = 0; i < size (); ++i)
    {
      bool unique = 0;

      for (object *op = spaces [i].bot; op; op = op->above)
        {
          unique |= op->flag [FLAG_UNIQUE] && op->flag [FLAG_IS_FLOOR];

          if (ecb_expect_false (!op->can_map_save ()))
            continue;

          if (ecb_expect_false (unique || op->flag [FLAG_UNIQUE]))
            {
              if (flags & IO_UNIQUES)
                op->write (f);
            }
          else if (ecb_expect_true (flags & IO_OBJECTS))
            op->write (f);
        }
    }

  coroapi::cede_to_tick ();

  return true;
}

bool
maptile::_save_objects (const char *path, int flags)
{
  object_freezer freezer;

  if (!_save_objects (freezer, flags))
    return false;

  return freezer.save (path);
}

void
maptile::init ()
{
  state = MAP_SWAPPED;

  /* The maps used to pick up default x and y values from the
   * map archetype. Mimic that behaviour.
   */
  width         = 16;
  height        = 16;
  timeout       = 300;
  max_items     = MAX_ITEM_PER_ACTION;
  max_volume    = 2000000; // 2m³
  reset_timeout = 0;
  enter_x       = 0;
  enter_y       = 0;
}

maptile::maptile ()
{
  init ();
}

maptile::maptile (int w, int h)
{
  init ();

  width  = w;
  height = h;

  alloc ();
}

/*
 * Allocates the arrays contained in a maptile.
 * This basically allocates the dynamic array of spaces for the
 * map.
 */
void
maptile::alloc ()
{
  if (spaces)
    return;

  spaces = salloc0<mapspace> (size ());
}

//+GPL

/* Takes a string from a map definition and outputs a pointer to the array of shopitems
 * corresponding to that string. Memory is allocated for this, it must be freed
 * at a later date.
 * Called by parse_map_headers below.
 */
static shopitems *
parse_shop_string (const char *input_string)
{
  char *shop_string, *p, *q, *next_semicolon, *next_colon;
  shopitems *items = NULL;
  int i = 0, number_of_entries = 0;
  const typedata *current_type;

  shop_string = strdup (input_string);
  p = shop_string;
  /* first we'll count the entries, we'll need that for allocating the array shortly */
  while (p)
    {
      p = strchr (p, ';');
      number_of_entries++;
      if (p)
        p++;
    }

  p = shop_string;
  strip_endline (p);
  items = new shopitems[number_of_entries + 1];
  for (i = 0; i < number_of_entries; i++)
    {
      if (!p)
        {
          LOG (llevError, "parse_shop_string: I seem to have run out of string, that shouldn't happen.\n");
          break;
        }

      next_semicolon = strchr (p, ';');
      next_colon = strchr (p, ':');
      /* if there is a stregth specified, figure out what it is, we'll need it soon. */
      if (next_colon && (!next_semicolon || next_colon < next_semicolon))
        items[i].strength = atoi (strchr (p, ':') + 1);

      if (isdigit (*p) || *p == '*')
        {
          items[i].typenum = atoi (p);  /* atoi returns 0 when we have an asterisk */
          current_type = get_typedata (items[i].typenum);
          if (current_type)
            {
              items[i].name = current_type->name;
              items[i].name_pl = current_type->name_pl;
            }
        }
      else
        {                       /*we have a named type, let's figure out what it is */
          q = strpbrk (p, ";:");
          if (q)
            *q = '\0';

          current_type = get_typedata_by_name (p);
          if (current_type)
            {
              items[i].name    = current_type->name;
              items[i].typenum = current_type->number;
              items[i].name_pl = current_type->name_pl;
            }
          else
            {                   /* oh uh, something's wrong, let's free up this one, and try
                                 * the next entry while we're at it, better print a warning
                                 */
              LOG (llevError, "invalid type %s defined in shopitems in string %s\n", p, input_string);
            }
        }

      items[i].index = number_of_entries;
      if (next_semicolon)
        p = ++next_semicolon;
      else
        p = NULL;
    }

  free (shop_string);
  return items;
}

/* opposite of parse string, this puts the string that was originally fed in to
 * the map (or something equivilent) into output_string. */
static const char *
print_shop_string (maptile *m)
{
  static dynbuf_text buf; buf.clear ();
  bool first = true;

  for (int i = 0; i < m->shopitems[0].index; i++)
    {
      if (!first)
        buf << ';';

      first = false;

      if (m->shopitems[i].typenum)
        {
          if (m->shopitems[i].strength)
            buf.printf ("%s:%d", m->shopitems[i].name, m->shopitems[i].strength);
          else
            buf.printf ("%s", m->shopitems[i].name);
        }
      else
        {
          if (m->shopitems[i].strength)
            buf.printf ("*:%d", m->shopitems[i].strength);
          else
            buf.printf ("*");
        }
    }

  return buf;
}

//-GPL

/* This loads the header information of the map.  The header
 * contains things like difficulty, size, timeout, etc.
 * this used to be stored in the map object, but with the
 * addition of tiling, fields beyond that easily named in an
 * object structure were needed, so it just made sense to
 * put all the stuff in the map object so that names actually make
 * sense.
 * This could be done in lex (like the object loader), but I think
 * currently, there are few enough fields this is not a big deal.
 * MSW 2001-07-01
 */
bool
maptile::_load_header (object_thawer &thawer)
{
  for (;;)
    {
      switch (thawer.kw)
        {
          case KW_msg:
            thawer.get_ml (KW_endmsg, msg);
            break;

          case KW_lore: // deliantra extension
            thawer.get_ml (KW_endlore, maplore);
            break;

          case KW_maplore:
            thawer.get_ml (KW_endmaplore, maplore);
            break;

          case KW_arch:
            if (strcmp (thawer.get_str (), "map"))
              LOG (llevError, "%s: loading map and got a non 'arch map' line (arch %s), skipping.\n", &path, thawer.get_str ());
            break;

          case KW_oid:
            thawer.get (this, thawer.get_sint32 ());
            break;

          case KW_file_format_version: break; // nop

          case KW_name:       thawer.get (name);       break;
          case KW_attach:     thawer.get (attach);     break;
          case KW_reset_time: thawer.get (reset_time); break;
          case KW_shopgreed:  thawer.get (shopgreed);  break;
          case KW_shopmin:    thawer.get (shopmin);    break;
          case KW_shopmax:    thawer.get (shopmax);    break;
          case KW_shoprace:   thawer.get (shoprace);   break;
          case KW_outdoor:    thawer.get (outdoor);    break;

          case KW_per_player: thawer.get (per_player); break;
          case KW_per_party:  thawer.get (per_party);  break;
          case KW_no_reset:   thawer.get (no_reset);   break;
          case KW_no_drop:    thawer.get (no_drop);    break;
                              
          case KW_region:     thawer.get (default_region); break;
          case KW_shopitems:  shopitems = parse_shop_string (thawer.get_str ()); break;

          //   old names            new names
          case KW_hp:          case KW_enter_x:         thawer.get (enter_x);  break;
          case KW_sp:          case KW_enter_y:         thawer.get (enter_y);  break;
          case KW_x:           case KW_width:           thawer.get (width);    break;
          case KW_y:           case KW_height:          thawer.get (height);   break;
          case KW_weight:      case KW_reset_timeout:   thawer.get (reset_timeout);   break;
          case KW_value:       case KW_swap_time:       thawer.get (timeout);  break;
          case KW_level:       case KW_difficulty:      thawer.get (difficulty); difficulty = clamp (difficulty, 1, settings.max_level); break;
          case KW_invisible:   case KW_darkness:        thawer.get (darkness); break;
          case KW_stand_still: case KW_fixed_resettime: thawer.get (fixed_resettime); break;

          case KW_tile_path_1: thawer.get (tile_path [TILE_NORTH]); break;
          case KW_tile_path_2: thawer.get (tile_path [TILE_EAST ]); break;
          case KW_tile_path_3: thawer.get (tile_path [TILE_SOUTH]); break;
          case KW_tile_path_4: thawer.get (tile_path [TILE_WEST ]); break;
          case KW_tile_path_5: thawer.get (tile_path [TILE_UP   ]); break;
          case KW_tile_path_6: thawer.get (tile_path [TILE_DOWN ]); break;

          case KW_ERROR:
            set_key_text (thawer.kw_str, thawer.value);
            break;

          case KW_end:
            thawer.next ();
            return true;

          default:
            if (!thawer.parse_error ("map"))
              return false;
            break;
        }

      thawer.next ();
    }

  abort ();
}

//+GPL

/******************************************************************************
 * This is the start of unique map handling code
 *****************************************************************************/

/* This goes through the maptile and removed any unique items on the map. */
void
maptile::clear_unique_items ()
{
  for (int i = 0; i < size (); ++i)
    {
      int unique = 0;
      for (object *op = spaces [i].bot; op; )
        {
          object *above = op->above;

          if (op->flag [FLAG_IS_FLOOR] && op->flag [FLAG_UNIQUE])
            unique = 1;

          if (op->head_ () == op && (op->flag [FLAG_UNIQUE] || unique))
            op->destroy ();

          op = above;
        }
    }
}

//-GPL

bool
maptile::_save_header (object_freezer &freezer)
{
#define MAP_OUT(k)    freezer.put (KW(k), k)
#define MAP_OUT2(k,v) freezer.put (KW(k), v)

  MAP_OUT2 (arch, CS(map));

  if (name) MAP_OUT (name);
  MAP_OUT (swap_time);
  MAP_OUT (reset_time);
  MAP_OUT (reset_timeout);
  MAP_OUT (fixed_resettime);
  MAP_OUT (no_reset);
  MAP_OUT (no_drop);
  MAP_OUT (difficulty);
  if (default_region) MAP_OUT2 (region, default_region->name);

  if (shopitems) MAP_OUT2 (shopitems, print_shop_string (this));
  MAP_OUT (shopgreed);
  MAP_OUT (shopmin);
  MAP_OUT (shopmax);
  if (shoprace) MAP_OUT (shoprace);

  MAP_OUT (width);
  MAP_OUT (height);
  MAP_OUT (enter_x);
  MAP_OUT (enter_y);
  MAP_OUT (darkness);
  MAP_OUT (outdoor);

  if (msg)     freezer.put (KW(msg)    , KW(endmsg)    , msg);
  if (maplore) freezer.put (KW(maplore), KW(endmaplore), maplore);

  MAP_OUT (per_player);
  MAP_OUT (per_party);

  if (tile_path [TILE_NORTH]) MAP_OUT2 (tile_path_1, tile_path [TILE_NORTH]);
  if (tile_path [TILE_EAST ]) MAP_OUT2 (tile_path_2, tile_path [TILE_EAST ]);
  if (tile_path [TILE_SOUTH]) MAP_OUT2 (tile_path_3, tile_path [TILE_SOUTH]);
  if (tile_path [TILE_WEST ]) MAP_OUT2 (tile_path_4, tile_path [TILE_WEST ]);
  if (tile_path [TILE_UP   ]) MAP_OUT2 (tile_path_5, tile_path [TILE_UP   ]);
  if (tile_path [TILE_DOWN ]) MAP_OUT2 (tile_path_6, tile_path [TILE_DOWN ]);

  freezer.put (this);
  freezer.put (KW(end));

  return true;
}

bool
maptile::_save_header (const char *path)
{
  object_freezer freezer;

  if (!_save_header (freezer))
    return false;

  return freezer.save (path);
}

//+GPL

/*
 * Remove and free all objects in the given map.
 */
void
maptile::clear ()
{
  if (spaces)
    {
      for (mapspace *ms = spaces + size (); ms-- > spaces; )
        while (object *op = ms->bot)
          {
            // manually remove, as to not trigger anything
            if (ms->bot = op->above)
              ms->bot->below = 0;

            op->flag [FLAG_REMOVED] = true;

            object *head = op->head_ ();
            if (op == head)
              op->destroy ();
            else if (head->map != op->map)
              {
                LOG (llevDebug, "bad luck for object crossing map borders: %s", head->debug_desc ());
                head->destroy ();
              }
          }

      sfree0 (spaces, size ());
    }

  if (buttons)
    free_objectlinkpt (buttons), buttons = 0;

  sfree0 (regions, size ());
  delete [] regionmap; regionmap = 0;
}

void
maptile::clear_header ()
{
  name     = 0;
  msg      = 0;
  maplore  = 0;
  shoprace = 0;
  delete [] shopitems, shopitems = 0;

  for (int i = 0; i < ecb_array_length (tile_path); i++)
    tile_path [i] = 0;
}

maptile::~maptile ()
{
  assert (destroyed ());
}

void
maptile::clear_links_to (maptile *m)
{
  /* We need to look through all the maps and see if any maps
   * are pointing at this one for tiling information.  Since
   * tiling can be asymetric, we just can not look to see which
   * maps this map tiles with and clears those.
   */
  for (int i = 0; i < ecb_array_length (tile_path); i++)
    if (tile_map[i] == m)
      tile_map[i] = 0;
}

void
maptile::do_destroy ()
{
  attachable::do_destroy ();

  clear ();
}

/* decay and destroy perishable items in a map */
// TODO: should be done regularly, not on map load?
void
maptile::do_decay_objects ()
{
  if (!spaces)
    return;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *above, *op = ms->bot; op; op = above)
      {
        above = op->above;

        // do not decay anything above unique floor tiles (yet :)
        if (op->flag [FLAG_IS_FLOOR] && op->flag [FLAG_UNIQUE])
          break;

        bool destroy = 0;

        if (op->flag [FLAG_IS_FLOOR]
            || op->flag [FLAG_OBJ_ORIGINAL]
            || op->flag [FLAG_UNIQUE]
            || op->flag [FLAG_OVERLAY_FLOOR]
            || op->flag [FLAG_UNPAID]
            || op->is_alive ())
          ; // do not decay
        else if (op->is_weapon ())
          {
            op->stats.dam--;
            if (op->stats.dam < 0)
              destroy = 1;
          }
        else if (op->is_armor ())
          {
            op->stats.ac--;
            if (op->stats.ac < 0)
              destroy = 1;
          }
        else if (op->type == FOOD)
          {
            op->stats.food -= rndm (5, 20);
            if (op->stats.food < 0)
              destroy = 1;
          }
        else
          {
            int mat = op->materials;

            if (mat & M_PAPER
                || mat & M_LEATHER
                || mat & M_WOOD
                || mat & M_ORGANIC
                || mat & M_CLOTH
                || mat & M_LIQUID
                || (mat & M_IRON  && rndm (1, 5) == 1)
                || (mat & M_GLASS && rndm (1, 2) == 1)
                || ((mat & M_STONE      || mat & M_ADAMANT) && rndm (1, 10) == 1)
                || ((mat & M_SOFT_METAL || mat & M_BONE)    && rndm (1, 3)  == 1)
                //|| (mat & M_ICE && temp > 32)
               )
              destroy = 1;
          }

        /* adjust overall chance below */
        if (destroy && rndm (0, 1))
          op->destroy ();
      }
}

/*
 * This routine is supposed to find out the difficulty of the map.
 * difficulty does not have a lot to do with character level,
 * but does have a lot to do with treasure on the map.
 *
 * Difficulty can now be set by the map creator. If the value stored
 * in the map is zero, then use this routine. Maps should really
 * have a difficulty set rather than using this function - human calculation
 * is much better than this function's guesswork.
 */
int
maptile::estimate_difficulty () const
{
  long monster_cnt = 0;
  double avgexp = 0;
  sint64 total_exp = 0;

  for (mapspace *ms = spaces + size (); ms-- > spaces; )
    for (object *op = ms->bot; op; op = op->above)
      {
        if (op->flag [FLAG_MONSTER])
          {
            total_exp += op->stats.exp;
            monster_cnt++;
          }

        if (op->flag [FLAG_GENERATOR])
          {
            total_exp += op->stats.exp;

            if (archetype *at = op->other_arch)
              {
                total_exp += at->stats.exp * 8;
                monster_cnt++;
              }

            for (object *inv = op->inv; inv; inv = inv->below)
              {
                total_exp += op->stats.exp * 8;
                monster_cnt++;
              }
          }
      }

  avgexp = (double) total_exp / monster_cnt;

  for (int i = 1; i <= settings.max_level; i++)
    if ((level_exp (i, 1) - level_exp (i - 1, 1)) > (100 * avgexp))
      return i;

  return 1;
}

/* change_map_light() - used to change map light level (darkness)
 * up or down. Returns true if successful.   It should now be
 * possible to change a value by more than 1.
 * Move this from los.c to map.c since this is more related
 * to maps than los.
 * postive values make it darker, negative make it brighter
 */
int
maptile::change_map_light (int change)
{
  /* Nothing to do */
  if (!change)
    return 0;

  /* inform all players on the map */
  if (change > 0)
    new_info_map (NDI_BLACK | NDI_UNIQUE, this, "It becomes darker.");
  else
    new_info_map (NDI_BLACK | NDI_UNIQUE, this, "It becomes brighter.");

  darkness = clamp (darkness + change, -MAX_DARKNESS, MAX_DARKNESS);

  /* All clients need to get re-updated for the change */
  update_all_map_los (this);

  return 1;
}

/*
 * This function updates various attributes about a specific space
 * on the map (what it looks like, whether it blocks magic,
 * has a living creatures, prevents people from passing
 * through, etc)
 */
void
mapspace::update_ ()
{
  object *last = 0;
  uint8 flags = P_UPTODATE;
  sint8 light = 0;
  MoveType move_block = 0, move_slow = 0, move_on = 0, move_off = 0, move_allow = 0;
  uint64_t volume = 0;
  uint32_t items = 0;
  object *anywhere = 0;
  uint8_t middle_visibility = 0;

  //object *middle = 0;
  //object *top    = 0;
  //object *floor  = 0;
  // this seems to generate better code than using locals, above
  object *&top    = faces_obj[0] = 0;
  object *&middle = faces_obj[1] = 0;
  object *&floor  = faces_obj[2] = 0;

  object::flags_t allflags; // all flags of all objects or'ed together

  for (object *tmp = bot; tmp; last = tmp, tmp = tmp->above)
    {
      // Lights are additive, up to MAX_LIGHT_RADIUS, see los.C)
      light += tmp->glow_radius;

      /* This call is needed in order to update objects the player
       * is standing in that have animations (ie, grass, fire, etc).
       * However, it also causes the look window to be re-drawn
       * 3 times each time the player moves, because many of the
       * functions the move_player calls eventualy call this.
       *
       * Always put the player down for drawing.
       */
      if (ecb_expect_true (!tmp->invisible))
        {
          if (ecb_expect_false (tmp->type == PLAYER || tmp->flag [FLAG_MONSTER]))
            top = tmp;
          else if (ecb_expect_false (tmp->flag [FLAG_IS_FLOOR]))
            {
              /* If we got a floor, that means middle and top were below it,
               * so should not be visible, so we clear them.
               */
              middle = 0;
              top    = 0;
              floor  = tmp;
              volume = 0;
              items  = 0;
            }
          else
            {
              if (ecb_expect_true (!tmp->flag [FLAG_NO_PICK]))
                {
                  ++items;
                  volume += tmp->volume ();
                }

              /* Flag anywhere have high priority */
              if (ecb_expect_false (tmp->flag [FLAG_SEE_ANYWHERE]))
                anywhere = tmp;

              /* Find the highest visible face around.  If equal
               * visibilities, we still want the one nearer to the
               * top
               */
              if (ecb_expect_false (::faces [tmp->face].visibility >= middle_visibility))
                {
                  middle_visibility = ::faces [tmp->face].visibility;
                  middle = tmp;
                }
            }
        }

      move_slow  |= tmp->move_slow;
      move_block |= tmp->move_block;
      move_on    |= tmp->move_on;
      move_off   |= tmp->move_off;
      move_allow |= tmp->move_allow;

      allflags |= tmp->flag;

      if (tmp->type == PLAYER)      flags |= P_PLAYER;
      if (tmp->type == SAFE_GROUND) flags |= P_SAFE;
    }

  // FLAG_SEE_ANYWHERE takes precedence
  if (anywhere)
    middle = anywhere;

  // ORing all flags together and checking them here is much faster
  if (allflags [FLAG_BLOCKSVIEW]) flags |= P_BLOCKSVIEW;
  if (allflags [FLAG_NO_MAGIC]  ) flags |= P_NO_MAGIC;
  if (allflags [FLAG_ALIVE]     ) flags |= P_IS_ALIVE;
  if (allflags [FLAG_DAMNED]    ) flags |= P_NO_CLERIC;

  this->light      = min (light, MAX_LIGHT_RADIUS);
  this->flags_     = flags;
  this->move_block = move_block & ~move_allow;
  this->move_on    = move_on;
  this->move_off   = move_off;
  this->move_slow  = move_slow;
  this->volume_    = (volume + 1023) / 1024;
  this->items_     = upos_min (items, 65535); // assume nrof <= 2**31

  /* At this point, we have a floor face (if there is a floor),
   * and the floor is set - we are not going to touch it at
   * this point.
   * middle contains the highest visibility face.
   * top contains a player/monster face, if there is one.
   *
   * We now need to fill in top.face and/or middle.face.
   */

  /* If the top face also happens to be high visibility, re-do our
   * middle face.  This should not happen, as we already have the
   * else statement above so middle should not get set.  OTOH, it
   * may be possible for the faces to match but be different objects.
   */
  if (top == middle)
    middle = 0;

  /* There are three posibilities at this point:
   * 1) top face is set, need middle to be set.
   * 2) middle is set, need to set top.
   * 3) neither middle or top is set - need to set both.
   */

  for (object *tmp = last; tmp; tmp = tmp->below)
    {
      /* Once we get to a floor, stop, since we already have a floor object */
      if (tmp->flag [FLAG_IS_FLOOR])
        break;

      /* If two top faces are already set, quit processing */
      if (top && middle)
        break;

      /* Only show visible faces */
      if (!tmp->invisible)
        {
          /* Fill in top if needed */
          if (!top)
            {
              top = tmp;
              if (top == middle)
                middle = 0;
            }
          else
            {
              /* top is already set - we should only get here if
               * middle is not set
               *
               * Set the middle face and break out, since there is nothing
               * more to fill in.  We don't check visiblity here, since
               *
               */
              if (tmp != top)
                {
                  middle = tmp;
                  break;
                }
            }
        }
    }

  if (middle == floor)
    middle = 0;

  if (top == middle)
    middle = 0;

  // set lower map transparent floor flag if applicable
  if (floor && floor->flag [FLAG_IS_TRANSPARENT_FLOOR] && !middle && !top)
    {
      floor->set_anim_frame (0);

      if (maptile *m = floor->map->tile_map [TILE_DOWN])
        {
          mapspace &ms = m->at (floor->x, floor->y);
          ms.update ();

          if (object *floor2 = ms.faces_obj [2])
            if (!floor2->flag [FLAG_IS_TRANSPARENT_FLOOR])
              {
                floor->set_anim_frame (1);
                top = floor;
                middle = ms.faces_obj [0];
                floor = floor2;
              }

          ms.pflags |= PF_VIS_UP;
        }
    }

#if 0
  faces_obj [0] = top;
  faces_obj [1] = middle;
  faces_obj [2] = floor;
#endif
}

void
mapspace::update_up ()
{
  // invalidate up
  if (!(pflags & PF_VIS_UP))
    return;

  pflags &= ~PF_VIS_UP;

  if (bot)
    if (maptile *m = bot->map->tile_map [TILE_UP])
      m->at (bot->x, bot->y).invalidate ();
}

maptile *
maptile::tile_available (int dir, bool load)
{
  // map is there and we don't need to load it OR it's loaded => return what we have
  if (tile_map [dir] && (!load || tile_map [dir]->linkable ()))
    return tile_map [dir];

  if (tile_path [dir])
    {
      // well, try to locate it then, if possible - maybe it's there already
      // this is the ONLY place in the server that links maps with each other,
      // so any kind of inter-map stuff has to be initiated here.
      if (maptile *m = find_async (tile_path [dir], this, load))
        {
          bool mismatch = false;

          if (dir == TILE_NORTH || dir == TILE_SOUTH || dir == TILE_UP || dir == TILE_DOWN)
            if (width != m->width)
              mismatch = true;

          if (dir == TILE_EAST  || dir == TILE_WEST  || dir == TILE_UP || dir == TILE_DOWN)
            if (height != m->height)
              mismatch = true;

          if (mismatch)
            {
              LOG (llevError, "tile dimension mismatch for direction %d from %s to %s\n",
                   dir, &path, &m->path);
              m = 0;
            }
          else if (0)//D
            {
              // as an optimisation, link us against the other map if the other map
              // has us as neighbour, which is very common, but not guaranteed.
              int dir2 = REVERSE_TILE_DIR (dir);

              if (m->tile_path [dir2] == path)
                m->tile_map [dir2] = this;
            }


          return tile_map [dir] = m;
        }
    }

  return 0;
}

/* this returns TRUE if the coordinates (x,y) are out of
 * map m.  This function also takes into account any
 * tiling considerations, loading adjacant maps as needed.
 * This is the function should always be used when it
 * necessary to check for valid coordinates.
 * This function will recursively call itself for the
 * tiled maps.
 */
int
out_of_map (maptile *m, int x, int y)
{
  /* If we get passed a null map, this is obviously the
   * case.  This generally shouldn't happen, but if the
   * map loads fail below, it could happen.
   */
  if (!m)
    return 0;

  if (x < 0)
    {
      if (!m->tile_available (TILE_WEST))
        return 1;

      return out_of_map (m->tile_map [TILE_WEST], x + m->tile_map [TILE_WEST]->width, y);
    }

  if (x >= m->width)
    {
      if (!m->tile_available (TILE_EAST))
        return 1;

      return out_of_map (m->tile_map [TILE_EAST], x - m->width, y);
    }

  if (y < 0)
    {
      if (!m->tile_available (TILE_NORTH))
        return 1;

      return out_of_map (m->tile_map [TILE_NORTH], x, y + m->tile_map [TILE_NORTH]->height);
    }

  if (y >= m->height)
    {
      if (!m->tile_available (TILE_SOUTH))
        return 1;

      return out_of_map (m->tile_map [TILE_SOUTH], x, y - m->height);
    }

  /* Simple case - coordinates are within this local
   * map.
   */
  return 0;
}

/* This is basically the same as out_of_map above, but
 * instead we return NULL if no map is valid (coordinates
 * out of bounds and no tiled map), otherwise it returns
 * the map as that the coordinates are really on, and
 * updates x and y to be the localised coordinates.
 * Using this is more efficient of calling out_of_map
 * and then figuring out what the real map is
 */
maptile *
maptile::xy_find (sint16 &x, sint16 &y)
{
  if (x < 0)
    {
      if (!tile_available (TILE_WEST))
        return 0;

      x += tile_map [TILE_WEST]->width;
      return tile_map [TILE_WEST]->xy_find (x, y);
    }

  if (x >= width)
    {
      if (!tile_available (TILE_EAST))
        return 0;

      x -= width;
      return tile_map [TILE_EAST]->xy_find (x, y);
    }

  if (y < 0)
    {
      if (!tile_available (TILE_NORTH))
        return 0;

      y += tile_map [TILE_NORTH]->height;
      return tile_map [TILE_NORTH]->xy_find (x, y);
    }

  if (y >= height)
    {
      if (!tile_available (TILE_SOUTH))
        return 0;

      y -= height;
      return tile_map [TILE_SOUTH]->xy_find (x, y);
    }

  /* Simple case - coordinates are within this local
   * map.
   */
  return this;
}

/**
 * Return whether map2 is adjacent to map1. If so, store the distance from
 * map1 to map2 in dx/dy.
 */
int
adjacent_map (maptile *map1, maptile *map2, int *dx, int *dy)
{
  if (!map1 || !map2)
    return 0;

  //TODO: this doesn't actually check correctly when intermediate maps are not loaded
  //fix: compare paths instead (this is likely faster, too!)
  if (map1 == map2)
    {
      *dx = 0;
      *dy = 0;
    }
  else if (map1->tile_available (TILE_NORTH, false) == map2)
    {
      *dx = 0;
      *dy = -map2->height;
    }
  else if (map1->tile_available (TILE_EAST , false) == map2)
    {
      *dx = map1->width;
      *dy = 0;
    }
  else if (map1->tile_available (TILE_SOUTH, false) == map2)
    {
      *dx = 0;
      *dy = map1->height;
    }
  else if (map1->tile_available (TILE_WEST , false) == map2)
    {
      *dx = -map2->width;
      *dy = 0;
    }
  else if (map1->tile_map[TILE_NORTH] && map1->tile_map[TILE_NORTH]->tile_available (TILE_EAST , false) == map2)
    {                           /* up right */
      *dx = +map1->tile_map[TILE_NORTH]->width;
      *dy = -map1->tile_map[TILE_NORTH]->height;
    }
  else if (map1->tile_map[TILE_NORTH] && map1->tile_map[TILE_NORTH]->tile_available (TILE_WEST , false) == map2)
    {                           /* up left */
      *dx = -map2->width;
      *dy = -map1->tile_map[TILE_NORTH]->height;
    }
  else if (map1->tile_map[TILE_EAST ] && map1->tile_map[TILE_EAST ]->tile_available (TILE_NORTH, false) == map2)
    {                           /* right up */
      *dx = +map1->width;
      *dy = -map2->height;
    }
  else if (map1->tile_map[TILE_EAST ] && map1->tile_map[TILE_EAST ]->tile_available (TILE_SOUTH, false) == map2)
    {                           /* right down */
      *dx = +map1->width;
      *dy = +map1->tile_map[TILE_EAST]->height;
    }
  else if (map1->tile_map[TILE_SOUTH] && map1->tile_map[TILE_SOUTH]->tile_available (TILE_EAST , false) == map2)
    {                           /* down right */
      *dx = +map1->tile_map[TILE_SOUTH]->width;
      *dy = +map1->height;
    }
  else if (map1->tile_map[TILE_SOUTH] && map1->tile_map[TILE_SOUTH]->tile_available (TILE_WEST , false) == map2)
    {                           /* down left */
      *dx = -map2->width;
      *dy = +map1->height;
    }
  else if (map1->tile_map[TILE_WEST ] && map1->tile_map[TILE_WEST ]->tile_available (TILE_NORTH, false) == map2)
    {                           /* left up */
      *dx = -map1->tile_map[TILE_WEST]->width;
      *dy = -map2->height;
    }
  else if (map1->tile_map[TILE_WEST ] && map1->tile_map[TILE_WEST ]->tile_available (TILE_SOUTH, false) == map2)
    {                           /* left down */
      *dx = -map1->tile_map[TILE_WEST]->width;
      *dy = +map1->tile_map[TILE_WEST]->height;
    }
  else
    return 0;

  return 1;
}

maptile *
maptile::xy_load (sint16 &x, sint16 &y)
{
  maptile *map = xy_find (x, y);

  if (map)
    map->load_sync ();

  return map;
}

maptile *
get_map_from_coord (maptile *m, sint16 *x, sint16 *y)
{
  return m->xy_load (*x, *y);
}

/* From map.c
 * This is used by get_player to determine where the other
 * creature is.  get_rangevector takes into account map tiling,
 * so you just can not look the the map coordinates and get the
 * righte value.  distance_x/y are distance away, which
 * can be negative.  direction is the crossfire direction scheme
 * that the creature should head.  part is the part of the
 * monster that is closest.
 *
 * get_rangevector looks at op1 and op2, and fills in the
 * structure for op1 to get to op2.
 * We already trust that the caller has verified that the
 * two objects are at least on adjacent maps.  If not,
 * results are not likely to be what is desired.
 * if the objects are not on maps, results are also likely to
 * be unexpected
 *
 * currently, the only flag supported (0x1) is don't translate for
 * closest body part of 'op1'
 */
void
get_rangevector (object *op1, object *op2, rv_vector *retval, int flags)
{
  if (!adjacent_map (op1->map, op2->map, &retval->distance_x, &retval->distance_y))
    {
      /* be conservative and fill in _some_ data */
      retval->distance   = 10000;
      retval->distance_x = 10000;
      retval->distance_y = 10000;
      retval->direction  = 0;
      retval->part       = 0;
    }
  else
    {
      retval->distance_x += op2->x - op1->x;
      retval->distance_y += op2->y - op1->y;

      object *best = op1;

      /* If this is multipart, find the closest part now */
      if (!(flags & 1) && op1->more)
        {
          int best_distance = idistance (retval->distance_x, retval->distance_y);

          /* we just take the offset of the piece to head to figure
           * distance instead of doing all that work above again
           * since the distance fields we set above are positive in the
           * same axis as is used for multipart objects, the simply arithmetic
           * below works.
           */
          for (object *tmp = op1->more; tmp; tmp = tmp->more)
            {
              int tmpi = idistance (op1->x - tmp->x + retval->distance_x, op1->y - tmp->y + retval->distance_y);

              if (tmpi < best_distance)
                {
                  best_distance = tmpi;
                  best          = tmp;
                }
            }

          if (best != op1)
            {
              retval->distance_x += op1->x - best->x;
              retval->distance_y += op1->y - best->y;
            }
        }

      retval->part      = best;
      retval->distance  = upos_max (abs (retval->distance_x), abs (retval->distance_y));
      retval->direction = find_dir_2 (retval->distance_x, retval->distance_y);
    }
}

/* this is basically the same as get_rangevector above, but instead of
 * the first parameter being an object, it instead is the map
 * and x,y coordinates - this is used for path to player -
 * since the object is not infact moving but we are trying to traverse
 * the path, we need this.
 * flags has no meaning for this function at this time - I kept it in to
 * be more consistant with the above function and also in case they are needed
 * for something in the future.  Also, since no object is pasted, the best
 * field of the rv_vector is set to NULL.
 */
void
get_rangevector_from_mapcoord (maptile *m, int x, int y, const object *op2, rv_vector *retval, int flags)
{
  if (!adjacent_map (m, op2->map, &retval->distance_x, &retval->distance_y))
    {
      /* be conservative and fill in _some_ data */
      retval->distance   = 100000;
      retval->distance_x = 32767;
      retval->distance_y = 32767;
      retval->direction  = 0;
      retval->part       = 0;
    }
  else
    {
      retval->distance_x += op2->x - x;
      retval->distance_y += op2->y - y;

      retval->part      = 0;
      retval->distance  = upos_max (abs (retval->distance_x), abs (retval->distance_y));
      retval->direction = find_dir_2 (retval->distance_x, retval->distance_y);
    }
}

/* Returns true of op1 and op2 are effectively on the same map
 * (as related to map tiling).  Note that this looks for a path from
 * op1 to op2, so if the tiled maps are asymetric and op2 has a path
 * to op1, this will still return false.
 * Note we only look one map out to keep the processing simple
 * and efficient.  This could probably be a macro.
 * MSW 2001-08-05
 */
int
on_same_map (const object *op1, const object *op2)
{
  int dx, dy;

  return adjacent_map (op1->map, op2->map, &dx, &dy);
}

//-GPL

object *
maptile::insert (object *op, int x, int y, object *originator, int flags)
{
  return insert_ob_in_map_at (op, this, originator, flags, x, y);
}

region *
maptile::region (int x, int y) const
{
  if (regions
      && regionmap
      && !OUT_OF_REAL_MAP (this, x, y))
    if (struct region *reg = regionmap [regions [y * width + x]])
      return reg;

  if (default_region)
    return default_region;

  return ::region::default_region ();
}

//+GPL

/* picks a random object from a style map.
 */
object *
maptile::pick_random_object (rand_gen &gen) const
{
  /* while returning a null object will result in a crash, that
   * is actually preferable to an infinite loop.  That is because
   * most servers will automatically restart in case of crash.
   * Change the logic on getting the random space - shouldn't make
   * any difference, but this seems clearer to me.
   */
  for (int i = 1000; --i;)
    {
      object *pick = at (gen (width), gen (height)).bot;

      // must be head: do not prefer big monsters just because they are big.
      if (pick && pick->is_head ())
        return pick;
    }

  // instead of crashing in the unlikely(?) case, try to return *something*
  return archetype::find (shstr_bug);
}

//-GPL

void
maptile::play_sound (faceidx sound, int x, int y) const
{
  if (!sound)
    return;

  for_all_players_on_map (pl, this)
    if (client *ns = pl->ns)
      {
        int dx = x - pl->ob->x;
        int dy = y - pl->ob->y;

        int distance = idistance (dx, dy);

        if (distance <= MAX_SOUND_DISTANCE)
          ns->play_sound (sound, dx, dy);
      }
}

void
maptile::say_msg (const char *msg, int x, int y) const
{
  for_all_players (pl)
    if (client *ns = pl->ns)
      {
        int dx = x - pl->ob->x;
        int dy = y - pl->ob->y;

        int distance = idistance (dx, dy);

        if (distance <= MAX_SOUND_DISTANCE)
          ns->send_msg (NDI_GREY | NDI_DEF, SAY_CHANNEL, msg);
      }
}

dynbuf mapwalk_buf (sizeof (maprect) * 25, sizeof (maprect) * 25);

static void
split_to_tiles (dynbuf &buf, maptile *m, int x0, int y0, int x1, int y1, int dx, int dy)
{
  // clip to map to the left
  if (x0 < 0)
    {
      if (maptile *tile = m->tile_available (TILE_WEST, 1))
        split_to_tiles (buf, tile, x0 + tile->width, y0, min (x1 + tile->width, tile->width), y1, dx - tile->width, dy);

      if (x1 < 0) // entirely to the left
        return;

      x0 = 0;
    }

  // clip to map to the right
  if (x1 > m->width)
    {
      if (maptile *tile = m->tile_available (TILE_EAST, 1))
        split_to_tiles (buf, tile, max (x0 - m->width, 0), y0, x1 - m->width, y1, dx + m->width, dy);

      if (x0 >= m->width) // entirely to the right
        return;

      x1 = m->width;
    }

  // clip to map to the north
  if (y0 < 0)
    {
      if (maptile *tile = m->tile_available (TILE_NORTH, 1))
        split_to_tiles (buf, tile, x0, y0 + tile->height, x1, min (y1 + tile->height, tile->height), dx, dy - tile->height);

      if (y1 < 0) // entirely to the north
        return;

      y0 = 0;
    }

  // clip to map to the south
  if (y1 > m->height)
    {
      if (maptile *tile = m->tile_available (TILE_SOUTH, 1))
        split_to_tiles (buf, tile, x0, max (y0 - m->height, 0), x1, y1 - m->height, dx, dy + m->height);

      if (y0 >= m->height) // entirely to the south
        return;

      y1 = m->height;
    }

  // if we get here, the rect is within the current map
  maprect *r = (maprect *)buf.alloc (sizeof (maprect));

  r->m  = m;
  r->x0 = x0;
  r->y0 = y0;
  r->x1 = x1;
  r->y1 = y1;
  r->dx = dx;
  r->dy = dy;
}

maprect *
maptile::split_to_tiles (dynbuf &buf, int x0, int y0, int x1, int y1)
{
  buf.clear ();

  ::split_to_tiles (buf, this, x0, y0, x1, y1, 0, 0);

  // add end marker
  maprect *r = (maprect *)buf.alloc (sizeof (maprect));
  r->m = 0;

  return (maprect *)buf.linearise ();
}

