/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2017,2018 Marc Alexander Lehmann / the Deliantra team
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992 Frank Tore Johansen
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

/* peterm@langmuir.eecs.berkeley.edu:  this function generates a random
snake-type maze.

input:  xsize, ysize;
output:  a char** array with # and . for closed and open respectively.

a char value of 0 represents a blank space:  a '#' is
a wall.

*/

#include <global.h>

#include <rmg.h>
#include "rproto.h"

/* These are some helper functions which help with
   manipulating a centered onion and turning it into
   a square spiral */
void
make_square_spiral_layout (layout &maze, int options)
{
  /* generate and allocate a doorless, centered onion */
  map_gen_onion (maze, RMOPT_CENTERED | RMOPT_NO_DOORS, 0);

  int xsize = maze.w;
  int ysize = maze.h;

  /* find the maze center.  */
  int cx = 0;
  int cy = 0;
  for (int i = 0; i < xsize; i++)
    for (int j = 0; j < ysize; j++)
      if (maze[i][j] == 'C')
        {
          cx = i;
          cy = j;
          break;
        }

  int tx = cx;
  int ty = cy;
  for (;;)
    {
      /* this starts from within a centered onion layer (or between two layers),
         and looks up until it finds a wall, and then looks right until it
         finds a vertical wall, i.e., the corner.  It sets tx and ty to that.
         it also starts from tx and ty. */
      --ty;

      /* find the top wall. */
      while (maze[tx][ty] != '#')
        --ty;

      /* proceed right until a corner is detected */
      while (maze[tx][ty + 1] != '#')
        ++tx;

      /* tx and ty should now be the top-right corner of the onion layer */
      if (ty < 2 || tx < 2 || tx > xsize - 2 || ty > ysize - 2)
        break;

      make_wall (maze, tx, ty - 1, 1);  /* make a vertical wall with a door */

      maze[tx][ty - 1] = '#';   /* convert the door that make_wall puts here to a wall */
      maze[tx - 1][ty] = 'D';   /* make a doorway out of this layer */

      /* walk left until we find the top-left corner */
      while (tx > 2 && maze[tx - 1][ty])
        --tx;

      make_wall (maze, tx - 1, ty, 0);  /* make a horizontal wall with a door */

      /* walk down until we find the bottom-left corner */
      while (ty + 1 < ysize && maze[tx][ty + 1])
        ++ty;

      make_wall (maze, tx, ty + 1, 1);  /* make a vertical wall with a door */

      /* walk rightuntil we find the bottom-right corner */
      while (tx + 1 < xsize && maze[tx + 1][ty])
        ++tx;

      make_wall (maze, tx + 1, ty, 0);  /* make a horizontal wall with a door */
      ++tx;                     /* set up for next layer. */
    }

  /* place the exits.  */

  if (rmg_rndm (2))
    {
      maze[cx][cy] = '>';
      maze[xsize - 2][1] = '<';
    }
  else
    {
      maze[cx][cy] = '<';
      maze[xsize - 2][1] = '>';
    }
}

