/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2017,2018 Marc Alexander Lehmann / the Deliantra team
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992 Frank Tore Johansen
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

/* Oct 3, 1995 - Code laid down for initial gods, priest alignment, and
 * monster race initialization. b.t.
 */

/* Sept 1996 - moved code over to object -oriented gods -b.t. */

#include <global.h>
#include <living.h>
#include <object.h>
#include <spells.h>
#include <sounds.h>
#include <sproto.h>

/**
 * Returns the id of specified god.
 */
static int
lookup_god_by_name (shstr_cmp name)
{
  if (name)
    for (godlink *gl = first_god; gl; gl = gl->next)
      if (gl->name == name)
        return gl->id;
  
  return -1;
}

/**
 * Returns pointer to specified god's object through pntr_to_god_obj..
 */
object *
find_god (shstr_cmp name)
{
  if (name)
    for (godlink *gl = first_god; gl; gl = gl->next)
      if (gl->name == name)
        return pntr_to_god_obj (gl);

  return 0;
}

/**
 * Returns a string that is the name of the god that should be natively worshipped by a
 * creature of who has race *race
 * if we can't find a god that is appropriate, we return NULL
 */
static shstr_tmp
get_god_for_race (shstr_cmp race)
{
  if (race)
    for (godlink *gl = first_god; gl; gl = gl->next)
      if (gl->arch->race == race)
        return gl->name;

  return shstr_tmp ();
}

/**
 * Determines if op worships a god.
 * Returns the godname if they do or "none" if they have no god.
 * In the case of an NPC, if they have no god, we try and guess
 * who they should worship based on their race. If that fails we
 * give them a random one.
 */
shstr_tmp
determine_god (object *op)
{
  /* spells */
  if ((op->type == SPELL || op->type == SPELL_EFFECT) && op->title)
    if (lookup_god_by_name (op->title) >= 0)
      return op->title;

  if (!op->is_player () && op->flag [FLAG_ALIVE])
    {
      /* find a god based on race */
      if (!op->title)
        op->title = get_god_for_race (op->race);

      /* find a random god */
      if (!op->title)
        {
          int godnr = rndm (1, first_god->id);

          for (godlink *gl = first_god; gl; gl = gl->next)
            if (gl->id == godnr)
              {
                op->title = gl->name;
                break;
              }
        }

      return op->title;
    }

  /* The god the player worships is in the praying skill (native skill
   * not skill tool).  Since a player can only have one instance of
   * that skill, once we find it, we can return, either with the
   * title or "none".
   */
  if (op->type == PLAYER)
    for (object *tmp = op->inv; tmp; tmp = tmp->below)
      if (tmp->type == SKILL && tmp->subtype == SK_PRAYING)
        {
          if (tmp->title)
            return tmp->title;

          break;
        }

  return shstr_none;
}

/**
 * Returns 1 if s1 and s2 are the same - either both NULL, or strcmp( ) == 0
 */
static int
same_string (const char *s1, const char *s2)
{
  if (s1 == s2)
    return 1;
  else if (s1 && s2)
    return !strcmp (s1, s2);
  else
    return 0;
}

/**
 * Checks for any occurrence of the given 'item' in the inventory of 'op' (recursively).
 * Any matching items in the inventory are deleted, and a
 * message is displayed to the player.
 */
static void
follower_remove_similar_item (object *op, object *item)
{
  object *tmp, *next;

  if (op && op->type == PLAYER && op->contr)
    {
      /* search the inventory */
      for (tmp = op->inv; tmp != NULL; tmp = next)
        {
          next = tmp->below;    /* backup in case we remove tmp */

          if (tmp->type == item->type
              && same_string (tmp->name, item->name)
              && same_string (tmp->title, item->title) && same_string (tmp->msg, item->msg) && same_string (tmp->slaying, item->slaying))
            {
              /* message */
              new_draw_info_format (NDI_UNIQUE, 0, op,
                                    tmp->nrof > 1 ? "The %s crumble to dust!" : "The %s crumbles to dust!",
                                    query_short_name (tmp));

              tmp->destroy ();
            }

          if (tmp->inv)
            follower_remove_similar_item (tmp, item);
        }
    }
}

/**
 * Checks for any occurrence of the given 'item' in the inventory of 'op' (recursively).
 * Returns 1 if found, else 0.
 */
static int
follower_has_similar_item (object *op, object *item)
{
  object *tmp;

  for (tmp = op->inv; tmp != NULL; tmp = tmp->below)
    {
      if (tmp->type == item->type
          && same_string (tmp->name, item->name)
          && same_string (tmp->title, item->title) && same_string (tmp->msg, item->msg) && same_string (tmp->slaying, item->slaying))
        return 1;
      if (tmp->inv && follower_has_similar_item (tmp, item))
        return 1;
    }
  return 0;
}

/**
 * God gives an item to the player.
 */
static int
god_gives_present (object *op, object *god, treasure *tr)
{
  if (!tr->item)
    return 0;

  if (follower_has_similar_item (op, tr->item))
    return 0;

  object *tmp = tr->item->instance ();
  new_draw_info_format (NDI_UNIQUE, 0, op, "%s lets %s appear in your hands.", &god->name, query_short_name (tmp));
  op->insert (tmp);

  return 1;
}

/**
 * Removes special prayers given by a god.
 */
static void
check_special_prayers (object *op, object *god)
{
  /* Ensure that 'op' doesn't know any special prayers that are not granted
   * by 'god'.
   */
  treasure *tr;
  object *tmp, *next_tmp;
  int remove = 0;

  /* Outer loop iterates over all special prayer marks */
  for (tmp = op->inv; tmp; tmp = next_tmp)
    {
      next_tmp = tmp->below;

      /* we mark special prayers with the STARTEQUIP flag, so if it isn't
       * in that category, not something we need to worry about.
       */
      if (tmp->type != SPELL || !tmp->flag [FLAG_STARTEQUIP])
        continue;

      if (god->randomitems == NULL)
        {
          LOG (llevError, "BUG: check_special_prayers(): god %s without randomitems\n", &god->name);
          do_forget_spell (op, tmp->name);
          continue;
        }

      /* Inner loop tries to find the special prayer in the god's treasure
       * list. We default that the spell should be removed.
       */
      remove = 1;
      for (tr = god->randomitems->items; tr; tr = tr->next)
        {
          object *item;

          if (!tr->item)
            continue;

          item = tr->item;

          /* Basically, see if the matching spell is granted by this god. */

          if (tr->item->type == SPELL && tr->item->object::name == tmp->name)
            {
              remove = 0;
              break;
            }
        }

      if (remove)
        do_forget_spell (op, tmp->name);
    }
}

/**
 * Unapplies up to number worth of items of type
 */
static void
stop_using_item (object *op, int type, int number)
{
  for (object *tmp = op->inv; tmp && number; tmp = tmp->below)
    if (tmp->type == type && tmp->flag [FLAG_APPLIED])
      {
        op->apply (tmp, AP_UNAPPLY | AP_IGNORE_CURSE);
        --number;
      }
}

/**
 * If the god does/doesnt have this flag, we
 * give/remove it from the experience object if it doesnt/does
 * already exist. For players only!
 */
static inline void
update_priest_flag (object *god, object *exp_ob, uint32 flag)
{
  exp_ob->flag [flag] = god->flag [flag];
  return;

  // old code follows for reference...
  if (god->flag [flag] && !exp_ob->flag [flag])
    exp_ob->set_flag (flag);
  else if (exp_ob->flag [flag] && !god->flag [flag])
    {
      /*  When this is called with the exp_ob set to the player,
       * this check is broken, because most all players arch
       * allow use of weapons.  I'm not actually sure why this
       * check is here - I guess if you had a case where the
       * value in the archetype (wisdom) should over ride the restrictions
       * the god places on it, this may make sense.  But I don't think
       * there is any case like that.
       */
/*        if (!(QUERY_FLAG(&(exp_ob->arch->clone),flag)))*/
      exp_ob->clr_flag (flag);
    };
}

/**
 * Forbids or let player use something item type.
 * op is the player.
 * exp_obj is the widsom experience.
 * flag is the flag to check against.
 * string is the string to print out.
 */
static int
worship_forbids_use (object *op, object *exp_obj, uint32 flag, const char *string)
{
  if (op->arch->flag [flag])
    if (op->flag [flag] != exp_obj->flag [flag])
      {
        update_priest_flag (exp_obj, op, flag);
        if (op->flag [flag])
          new_draw_info_format (NDI_UNIQUE, 0, op, "You may use %s again.", string);
        else
          {
            new_draw_info_format (NDI_UNIQUE, 0, op, "You are forbidden to use %s.", string);
            return 1;
          }
      }

  return 0;
}

/**
 * This function is called whenever a player has
 * switched to a new god. It handles basically all the stat changes
 * that happen to the player, including the removal of godgiven
 * items (from the former cult).
 */
void
object::become_follower (object *new_god)
{
  object *old_god = 0;       /* old god */
  treasure *tr;
  object *item, *skop, *next;
  int i, sk_applied, undeadified = 0;   /* Turns to true if changing god can changes the undead status of the player. */

  if (!contr)
    return;

  contr->queue_stats_update ();

  old_god = find_god (determine_god (this));

  /* take away any special god-characteristic items. */
  for (item = inv; item; item = next)
    {
      next = item->below;

      // remove all invisible startequip items which are not skill, exp or force
      if (item->flag [FLAG_STARTEQUIP] && item->invisible
          && item->type != SKILL && item->type != FORCE)
        {
          if (item->type == SPELL)
            {
              new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, this, "You lose knowledge of %s.", &item->name);
              esrv_remove_spell (contr, item);
            }

          item->destroy ();
        }
    }

  /* remove any godgiven items from the old god */
  if (old_god)
    for (tr = old_god->randomitems->items; tr; tr = tr->next)
      if (tr->item && tr->item->flag [FLAG_STARTEQUIP])
        follower_remove_similar_item (this, tr->item);

  if (!new_god)
    return;

  if (new_god->slaying && new_god->slaying.contains (race))
    {
      failmsgf ("Fool! %s detests your kind! H<This god does not accept followers of your race.>", &new_god->name);

      if (random_roll (0, level - 1, this, PREFER_LOW) - 5 > 0)
        {
          object *tmp = archetype::get (LOOSE_MANA);

          cast_magic_storm (this, tmp, new_god->level + 10);
        }

      return;
    }

  /* give the player any special god-characteristic-items. */
  for (tr = new_god->randomitems->items; tr; tr = tr->next)
    if (tr->item && tr->item->invisible && tr->item->type != SPELLBOOK
        && tr->item->type != BOOK && tr->item->type != SPELL)
      god_gives_present (this, new_god, tr);

  new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, this, "You become a follower of %s!", &new_god->name);

  for (skop = inv; skop; skop = skop->below)
    if (skop->type == SKILL && skop->subtype == SK_PRAYING)
      break;

  /* Player has no skill - give them the skill */
  if (!skop)
    /* The archetype should always be defined - if we crash here because it doesn't,
     * things are really messed up anyways.
     */
    skop = give_skill_by_name (this, shstr_praying);

  sk_applied = skop->flag [FLAG_APPLIED]; /* save skill status */

  /* Clear the "undead" status. We also need to force a call to change_abil,
   * so I set undeadified for that.
   * - gros, 21th July 2006.
   */
  if (old_god && old_god->flag [FLAG_UNDEAD])
    {
      skop->clr_flag (FLAG_UNDEAD);
      undeadified = 1;
    }

  if (skop->title)
    {
      /* get rid of old god */
      new_draw_info_format (NDI_UNIQUE, 0, this, "%s's blessing is withdrawn from you.", &skop->title);

      /* The point of this is to really show what abilities the player just lost */
      if (sk_applied || undeadified)
        {
          skop->clr_flag (FLAG_APPLIED);
          change_abil (this, skop);
        }
    }

  /* now change to the new gods attributes to exp_obj */
  skop->title = new_god->name;
  skop->path_attuned = new_god->path_attuned;
  skop->path_repelled = new_god->path_repelled;
  skop->path_denied = new_god->path_denied;
  /* copy god's resistances */
  memcpy (skop->resist, new_god->resist, sizeof (new_god->resist));

  /* make sure that certain immunities do NOT get passed
   * to the follower!
   */
  for (i = 0; i < NROFATTACKS; i++)
    if (skop->resist[i] > 30 && (i == ATNR_FIRE || i == ATNR_COLD || i == ATNR_ELECTRICITY || i == ATNR_POISON))
      skop->resist[i] = 30;

  skop->stats.hp = (sint16) new_god->last_heal;
  skop->stats.sp = (sint16) new_god->last_sp;
  skop->stats.grace = (sint16) new_god->last_grace;
  skop->stats.food = (sint16) new_god->last_eat;
  skop->stats.luck = (sint8) new_god->stats.luck;
  /* gods may pass on certain flag properties */
  update_priest_flag (new_god, skop, FLAG_SEE_IN_DARK);
  update_priest_flag (new_god, skop, FLAG_REFL_SPELL);
  update_priest_flag (new_god, skop, FLAG_REFL_MISSILE);
  update_priest_flag (new_god, skop, FLAG_STEALTH);
  update_priest_flag (new_god, skop, FLAG_MAKE_INVIS);
  update_priest_flag (new_god, skop, FLAG_UNDEAD);
  update_priest_flag (new_god, skop, FLAG_BLIND);
  update_priest_flag (new_god, skop, FLAG_XRAYS);       /* better have this if blind! */

  new_draw_info_format (NDI_UNIQUE, 0, this, "You are bathed in %s's aura.", &new_god->name);

  /* Weapon/armour use are special...handle flag toggles here as this can
   * only happen when gods are worshipped and if the new priest could
   * have used armour/weapons in the first place.
   *
   * This also can happen for monks which cannot use weapons. In this case
   * do not allow to use weapons even if the god otherwise would allow it.
   */
  if (!present_in_ob_by_name (FORCE, "no weapon force", this))
    update_priest_flag (new_god, skop, FLAG_USE_WEAPON);

  update_priest_flag (new_god, skop, FLAG_USE_ARMOUR);

  if (worship_forbids_use (this, skop, FLAG_USE_WEAPON, "weapons"))
    stop_using_item (this, WEAPON, 2);

  if (worship_forbids_use (this, skop, FLAG_USE_ARMOUR, "armour"))
    {
      stop_using_item (this, ARMOUR, 1);
      stop_using_item (this, HELMET, 1);
      stop_using_item (this, BOOTS, 1);
      stop_using_item (this, GLOVES, 1);
      stop_using_item (this, SHIELD, 1);
    }

  skop->set_flag (FLAG_APPLIED);
  change_abil (this, skop);

  /* return to previous skill status */
  if (!sk_applied)
    {
      skop->clr_flag (FLAG_APPLIED);
      contr->queue_stats_update ();
    }

  check_special_prayers (this, new_god);
}

archetype *
determine_holy_arch (object *god, shstr_cmp type)
{
  treasure *tr;

  if (!god || !god->randomitems)
    {
      LOG (llevError, "BUG: determine_holy_arch(): no god or god without " "randomitems\n");
      return 0;
    }

  for (tr = god->randomitems->items; tr; tr = tr->next)
    {
      if (!tr->item)
        continue;

      object *item = tr->item;

      if (item->type == BOOK && item->invisible && item->name == type)
        return item->other_arch;
    }

  return 0;
}

/**
 * God helps player by removing curse and/or damnation.
 */
static int
god_removes_curse (object *op, int remove_damnation)
{
  int success = 0;

  for (object *tmp = op->inv; tmp; tmp = tmp->below)
    {
      if (tmp->invisible)
        continue;

      if (tmp->flag [FLAG_DAMNED] && !remove_damnation)
        continue;

      if (tmp->flag [FLAG_CURSED] || tmp->flag [FLAG_DAMNED])
        {
          success = 1;
          tmp->clr_flag (FLAG_DAMNED);
          tmp->clr_flag (FLAG_CURSED);
          tmp->clr_flag (FLAG_KNOWN_CURSED);

          if (object *pl = tmp->visible_to ())
            esrv_update_item (UPD_FLAGS, pl, tmp);
        }
    }

  if (success)
    new_draw_info (NDI_UNIQUE, 0, op, "You feel like someone is helping you.");

  return success;
}

static int
follower_level_to_enchantments (int level, int difficulty)
{
  if (difficulty < 1)
    {
      LOG (llevError, "follower_level_to_enchantments(): " "difficulty %d is invalid\n", difficulty);
      return 0;
    }

  if (level <= 20)
    return level / difficulty;

  if (level <= 40)
    return (20 + (level - 20) / 2) / difficulty;

  return (30 + (level - 40) / 4) / difficulty;
}

/**
 * God checks item the player is using.
 * Return either -1 (bad), 0 (neutral) or
 * 1 (item is ok). If you are using the item of an enemy
 * god, it can be bad...-b.t.
 */
static int
god_examines_item (object *god, object *item)
{
  char buf[MAX_BUF];

  if (!god || !item)
    return 0;

  if (!item->title)
    return 1;                   /* unclaimed item are ok */

  sprintf (buf, "of %s", &god->name);
  if (!strcmp (&item->title, buf))
    return 1;                   /* belongs to that God */

  if (god->title)
    {                           /* check if we have any enemy blessed item */
      sprintf (buf, "of %s", &god->title);
      if (!strcmp (&item->title, buf))
        {
          if (item->env)
            {
              char buf[MAX_BUF];

              sprintf (buf, "Heretic! You are using %s!", query_name (item));
              new_draw_info (NDI_UNIQUE | NDI_NAVY, 0, item->env, buf);
            }

          return -1;
        }
    }

  return 0;                     /* item is sacred to a non-enemy god/or is otherwise magical */
}

/**
 * God wants to enchant weapon.
 * Affected weapon is the applied one (weapon or bow). It's checked to make sure
 * it isn't a weapon for another god. If all is all right, update weapon with
 * attacktype, slaying and such.
 */
static int
god_enchants_weapon (object *op, object *god, object *tr, object *skill)
{
  if (!op->contr)
    return 0;

  object *weapon = op->contr->combat_ob;

  if (!weapon)
    return 0;

  if (weapon->type != WEAPON && weapon->type != BOW)
    return 0;

  if (god_examines_item (god, weapon) <= 0)
    return 0;

  /* First give it a title, so other gods won't touch it */
  if (!weapon->title)
    {
      weapon->title = format ("of %s", &god->name);

      if (object *pl = weapon->visible_to ())
        esrv_update_item (UPD_NAME, pl, weapon);

      new_draw_info (NDI_UNIQUE, 0, op, "Your weapon quivers as if struck!");
    }

  /* Allow the weapon to slay enemies */
  if (!weapon->slaying && god->slaying)
    {
      weapon->slaying = god->slaying;
      new_draw_info_format (NDI_UNIQUE, 0, op, "Your %s now hungers to slay enemies of your god!", &weapon->name);
      return 1;
    }

  /* Add the gods attacktype */
  uint32 attacktype = (weapon->attacktype == 0) ? AT_PHYSICAL : weapon->attacktype;

  if ((attacktype & god->attacktype) != god->attacktype)
    {
      new_draw_info (NDI_UNIQUE, 0, op, "Your weapon suddenly glows!");
      weapon->attacktype = attacktype | god->attacktype;
      return 1;
    }

  /* Higher magic value */
  if (weapon->magic < follower_level_to_enchantments (skill->level, tr->level))
    {
      new_draw_info (NDI_UNIQUE, 0, op, "A phosphorescent glow envelops your weapon!");
      weapon->magic++;

      if (object *pl = weapon->visible_to ())
        esrv_update_item (UPD_NAME, pl, weapon);

      return 1;
    }

  return 0;
}

/**
 * Checks and maybe punishes someone praying.
 * All applied items are examined, if player is using more items of other gods,
 * s/he loses experience in praying or general experience if no praying.
 */
static int
god_examines_priest (object *op, object *god)
{
  int reaction = 1;
  object *item = NULL, *skop;

  for (item = op->inv; item; item = item->below)
    if (item->flag [FLAG_APPLIED])
      reaction += god_examines_item (god, item) * (item->magic ? abs (item->magic) : 1);

  /* well, well. Looks like we screwed up. Time for god's revenge */
  if (reaction < 0)
    {
      int loss = 10000000;
      int angry = abs (reaction);

      for (skop = op->inv; skop; skop = skop->below)
        if (skop->type == SKILL && skop->subtype == SK_PRAYING)
          break;

      if (skop)
        loss = 0.05f * skop->stats.exp;

      change_exp (op, -random_roll (0, loss * angry - 1, op, PREFER_LOW), skop ? skop->skill : shstr_none, SK_SUBTRACT_SKILL_EXP);

      if (random_roll (0, angry, op, PREFER_LOW))
        {
          object *tmp = archetype::get (LOOSE_MANA);

          cast_magic_storm (op, tmp, op->level + (angry * 3));
        }

      new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, op, "%s becomes angry and punishes you!", &god->name);
    }

  return reaction;
}

/**
 * Every once in a while the god will intervene to help the worshiper.
 * Later, this fctn can be used to supply quests, etc for the
 * priest. -b.t.
 * called from pray_at_altar() currently.
 */
static void
god_intervention (object *op, object *god, object *skill)
{
  treasure *tr;

  if (!god || !god->randomitems)
    {
      LOG (llevError, "BUG: god_intervention(): no god or god without randomitems\n");
      return;
    }

  check_special_prayers (op, god);

  /* lets do some checks of whether we are kosher with our god */
  if (god_examines_priest (op, god) < 0)
    return;

  op->play_sound (sound_find ("god_intervention"));
  new_draw_info (NDI_UNIQUE, 0, op, "You feel a holy presence!");

  for (tr = god->randomitems->items; tr; tr = tr->next)
    {
      object *item;

      if (tr->chance <= random_roll (0, 99, op, PREFER_HIGH))
        continue;

      /* Treasurelist - generate some treasure for the follower */
      if (tr->name)
        {
          treasurelist *tl = treasurelist::find (tr->name);

          if (tl == NULL)
            continue;

          new_draw_info (NDI_UNIQUE, 0, op, "Something appears before your eyes. You catch it before it falls to the ground.");

          create_treasure (tl, op, GT_STARTEQUIP | GT_ONLY_GOOD, skill->level, 0);
          return;
        }

      if (!tr->item)
        continue;

      item = tr->item;

      /* Grace limit */
      if (item->type == BOOK && item->invisible && item->name == shstr_grace_limit)
        {
          if (op->stats.grace < item->stats.grace || op->stats.grace < op->stats.maxgrace)
            {
              /* Follower lacks the required grace for the following
               * treasure list items. */

              object *tmp = archetype::get (HOLY_POSSESSION);
              cast_change_ability (op, op, tmp, 0, 1);
              tmp->destroy ();
              return;
            }

          continue;
        }

      /* Restore grace */
      if (item->type == BOOK && item->invisible && item->name == shstr_restore_grace)
        {
          if (op->stats.grace >= 0)
            continue;

          op->stats.grace = random_roll (0, 9, op, PREFER_HIGH);
          new_draw_info (NDI_UNIQUE, 0, op, "You are returned to a state of grace.");
          return;
        }

      /* Heal damage */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "restore hitpoints") == 0)
        {
          if (op->stats.hp >= op->stats.maxhp)
            continue;

          new_draw_info (NDI_UNIQUE, 0, op, "A white light surrounds and heals you!");
          op->stats.hp = op->stats.maxhp;
          return;
        }

      /* Restore spellpoints */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "restore spellpoints") == 0)
        {
          int max = (int) (op->stats.maxsp * (item->stats.maxsp / 100.0));

          /* Restore to 50 .. 100%, if sp < 50% */
          int new_sp = (int) (random_roll (1000, 1999, op, PREFER_HIGH) / 2000.0 * max);

          if (op->stats.sp >= max / 2)
            continue;

          new_draw_info (NDI_UNIQUE, 0, op, "A blue lightning strikes your head but doesn't hurt you!");
          op->stats.sp = new_sp;
        }

      /* Various heal spells */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "heal spell") == 0)
        {
          object *tmp = archetype::get (item->slaying);
          int success = cast_heal (op, op, tmp, 0);
          tmp->destroy ();

          if (success)
            return;
          else
            continue;
        }

      /* Remove curse */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "remove curse") == 0)
        {
          if (god_removes_curse (op, 0))
            return;
          else
            continue;
        }

      /* Remove damnation */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "remove damnation") == 0)
        {
          if (god_removes_curse (op, 1))
            return;
          else
            continue;
        }

      /* Heal depletion */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "heal depletion") == 0)
        {
          object *depl;
          archetype *at;
          int i;

          if ((at = archetype::find (shstr_depletion)) == NULL)
            {
              LOG (llevError, "Could not find archetype depletion.\n");
              continue;
            }

          depl = present_arch_in_ob (at, op);

          if (depl == NULL)
            continue;

          new_draw_info (NDI_UNIQUE, 0, op, "Shimmering light surrounds and restores you!");

          for (i = 0; i < NUM_STATS; i++)
            if (depl->stats.stat (i))
              new_draw_info (NDI_UNIQUE, 0, op, restore_msg[i]);

          depl->destroy ();
          op->update_stats ();
          return;
        }

      /* Voices */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "voice_behind") == 0)
        {
          new_draw_info (NDI_UNIQUE, 0, op, "You hear a voice from behind you, but you don't dare to turn around:");
          new_draw_info (NDI_WHITE, 0, op, item->msg);
          return;
        }

      /* Messages */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "message") == 0)
        {
          new_draw_info (NDI_UNIQUE, 0, op, item->msg);
          return;
        }

      /* Enchant weapon */
      if (item->type == BOOK && item->invisible && strcmp (item->name, "enchant weapon") == 0)
        {
          if (god_enchants_weapon (op, god, item, skill))
            return;
          else
            continue;
        }

      /* Spellbooks - works correctly only for prayers */
      if (item->type == SPELL)
        {
          if (check_spell_known (op, item->name))
            continue;
          if (item->level > skill->level)
            continue;

          new_draw_info_format (NDI_UNIQUE, 0, op, "%s grants you use of a special prayer!", &god->name);
          do_learn_spell (op, item, 1);
          return;

        }

      /* Other gifts */
      if (!item->invisible)
        {
          if (god_gives_present (op, god, tr))
            return;
          else
            continue;
        }
      /* else ignore it */
    }

  new_draw_info (NDI_UNIQUE, 0, op, "You feel rapture.");
}

/**
 * Player prays at altar.
 * Checks for god changing, divine intervention, and so on.
 */
void
pray_at_altar (object *pl, object *altar, object *skill)
{
  object *pl_god = find_god (determine_god (pl));

  if (INVOKE_PLAYER (PRAY_ALTAR, pl->contr, ARG_OBJECT (altar), ARG_OBJECT (skill)))
    return;

  /* If non consecrate altar, don't do anything */
  if (!altar->other_arch)
    return;

  /* hmm. what happend depends on pl's current god, level, etc */
  if (!pl_god)
    {                           /*new convert */
      pl->become_follower (altar->other_arch);
      return;
    }
  else if (pl_god->name == altar->other_arch->object::name)
    {
      /* pray at your gods altar */
      /* this leads to very low levels of wis and pray to result in no doubling! */
      int bonus = (pl->stats.Wis + skill->level) / 10;

      /* we can get neg grace up faster */
      if (pl->stats.grace < 0)
        pl->stats.grace += (bonus > -1 * (pl->stats.grace / 10) ? bonus : -1 * (pl->stats.grace / 10));

      /* we can super-charge grace to 2x max */
      if (pl->stats.grace < 2 * pl->stats.maxgrace)
        pl->stats.grace += bonus / 2;
      else
        pl->stats.grace = 2 * pl->stats.maxgrace;

      /* Every once in a while, the god decides to checkup on their
       * follower, and may intervene to help them out.
       */
      bonus = max (1, bonus + max (pl->stats.luck, -3));        /* -- DAMN -- */

      if (((random_roll (0, 399, pl, PREFER_LOW)) - bonus) < 0)
        god_intervention (pl, pl_god, skill);
    }
  else
    {                           /* praying to another god! */
      uint64 loss = 0;
      int angry = 1;

      /* I believe the logic for detecting opposing gods was completely
       * broken - I think it should work now.  altar->other_arch
       * points to the god of this altar (which we have
       * already verified is non null).  pl_god->other_arch
       * is the opposing god - we need to verify that exists before
       * using its values.
       */
      if (pl_god->other_arch && (altar->other_arch->archname == pl_god->other_arch->archname))
        {
          angry = 2;
          if (random_roll (0, skill->level + 2, pl, PREFER_LOW) - 5 > 0)
            {
              object *tmp;

              /* you really screwed up */
              angry = 3;
              new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, pl, "Foul Priest! %s punishes you!", &pl_god->name);
              tmp = archetype::get (LOOSE_MANA);
              cast_magic_storm (pl, tmp, pl_god->level + 20);
            }
          else
            new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, pl, "Foolish heretic! %s is livid!", &pl_god->name);
        }
      else
        new_draw_info_format (NDI_UNIQUE | NDI_NAVY, 0, pl, "Heretic! %s is angered!", &pl_god->name);

      /* whether we will be successfull in defecting or not -
       * we lose experience from the clerical experience obj
       */

      loss = angry * (skill->stats.exp / 10);
      if (loss)
        change_exp (pl, -random_roll64 (0, loss, pl, PREFER_LOW), skill ? skill->skill : shstr_none, SK_SUBTRACT_SKILL_EXP);

      /* May switch Gods, but its random chance based on our current level
       * note it gets harder to swap gods the higher we get
       */
      if ((angry == 1) && !(random_roll (0, skill->level, pl, PREFER_LOW)))
        pl->become_follower (altar->other_arch);
      else
        {
          /* toss this player off the altar.  He can try again. */
          new_draw_info (NDI_UNIQUE | NDI_NAVY, 0, pl, "A divine force pushes you off the altar.");
          pl->contr->fire_on = 0;
          pl->speed_left = 1.f;
          move_player (pl, absdir (pl->facing + 4));    /* back him off the way he came. */
        }
    }
}

/**
 * Changes the attributes of cone, smite, and ball spells as needed by the code.
 * Returns false if there was no race to assign to the slaying field of the spell, but
 * the spell attacktype contains AT_HOLYWORD.  -b.t.
 */
int
tailor_god_spell (object *spellop, object *caster)
{
  object *god = find_god (determine_god (caster));
  int caster_is_spell = 0;

  if (caster->type == SPELL_EFFECT || caster->type == SPELL)
    caster_is_spell = 1;

  /* if caster is a rune or the like, it doesn't worship anything.  However,
   * if this object is owned by someone, then the god that they worship
   * is relevant, so use that.
   */
  if (!god && caster->owner)
    god = find_god (determine_god (caster->owner));

  if (!god || (spellop->attacktype & AT_HOLYWORD && !god->race))
    {
      if (!caster_is_spell)
        new_draw_info (NDI_UNIQUE, 0, caster, "This prayer is useless unless you worship an appropriate god");
      else
        LOG (llevError, "BUG: tailor_god_spell(): no god\n");

      spellop->destroy ();
      return 0;
    }

  /* either holy word or godpower attacks will set the slaying field */
  if (spellop->attacktype & AT_HOLYWORD || spellop->attacktype & AT_GODPOWER)
    {
      if (spellop->slaying)
        spellop->slaying = NULL;

      if (!caster_is_spell)
        spellop->slaying = god->slaying;
      else if (caster->slaying)
        spellop->slaying = caster->slaying;
    }

  /* only the godpower attacktype adds the god's attack onto the spell */
  if (spellop->attacktype & AT_GODPOWER)
    spellop->attacktype = spellop->attacktype | god->attacktype;

  /* tack on the god's name to the spell */
  if (spellop->attacktype & AT_HOLYWORD || spellop->attacktype & AT_GODPOWER)
    {
      spellop->title = god->name;
      if (spellop->title)
        {
          char buf[MAX_BUF];

          sprintf (buf, "%s of %s", &spellop->name, &spellop->title);
          spellop->name = spellop->name_pl = buf;
        }
    }

  return 1;
}
