/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2018 Marc Alexander Lehmann / the Deliantra team
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <global.h>
#include <sproto.h>
#include <cstdarg>

#if HAVE_TCP_INFO
# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <netinet/tcp.h>
#endif

using namespace std;

// force a packet when idle for more than this many seconds,
// forcing an ack regularly.
#define IDLE_PING 2.

void
client::flush ()
{
  if (destroyed ())
    return;

  /**
   * Writes data to socket.
   *
   * When the socket is clear to write, and we have backlogged data, this
   * is called to write it out.
   */

  // write a nop to the socket at least every IDLE_PING seconds.
  if (!outputbuffer.len)
    {
      if (last_send + IDLE_PING <= NOW && pl && pl->active)
        {
          // this is a bit ugly, but map1a seems to be the only
          // nop'able command and it is quite small.
          packet sl ("map1a");
          send_packet (sl);
        }
      else
        return;
    }

  if (socket_ev.poll () & EV_WRITE)
    return;

  last_send = NOW;
  write_outputbuffer ();
}

void
client::write_outputbuffer ()
{
  while (outputbuffer.len)
    {
      int res = write (fd, outputbuffer.data + outputbuffer.start,
                       min (outputbuffer.len, SOCKETBUFSIZE - outputbuffer.start));

      if (res > 0)
        {
          outputbuffer.start += res;
          /* wrap back to start of buffer */
          if (outputbuffer.start == SOCKETBUFSIZE)
            outputbuffer.start = 0;

          outputbuffer.len -= res;
        }
      else if (res == 0)
        {
          LOG (llevError, "socket write failed, connection closed.\n");
          destroy ();
          return;
        }
      else if (errno == EINTR)
        {
          // just retry
        }
      else if (errno == EAGAIN)
        {
          // delay till ready
          socket_ev.poll (socket_ev.poll () | EV_WRITE);
          socket_ev.start ();
          return;
        }
      else
        {
          LOG (llevError, "socket write failed: %s\n", strerror (errno));
          destroy ();
          return;
        }
    }

  socket_ev.poll (socket_ev.poll () & ~EV_WRITE);
}

/******************************************************************************
 *
 * Start of read routines.
 *
 ******************************************************************************/

int
client::next_packet ()
{
  if (inbuf_len >= 2)
    {
      int pkt_len = (inbuf [0] << 8) | inbuf [1];

      if (inbuf_len >= 2 + pkt_len)
        return 2 + pkt_len;

      if (inbuf_len == sizeof (inbuf))
        {
          send_packet_printf ("msg %d log input buffer overflow - closing connection.", NDI_RED | NDI_REPLY);
          destroy ();
          return -1;
        }
    }

  return 0;
}

void
client::skip_packet (int len)
{
  inbuf_len -= len;
  memmove (inbuf, inbuf + len, inbuf_len);
}

/*****************************************************************************
 * Start of command dispatch area.
 * The commands here are protocol commands.
 ****************************************************************************/

// SocketCommand, PlayingCommand, should not exist with those ugly casts
#define SC(cb) (void *)static_cast<void (*)(char *, int, client *)>(cb),
#define PC(cb) (void *)static_cast<void (*)(char *, int, player *)>(cb), PF_PLAYER |

static void
NopCmd (char *, int, client *)
{
  // do nothing
}

/**
 * Dispatch table for the server.
 */
static struct packet_type packets[] = {
  {"command",     PC(PlayerCmd)    PF_PLAYING | PF_COMMAND },

  {"examine",     PC(ExamineCmd)   PF_PLAYING },
  {"ex",          PC(ExCmd)        PF_PLAYING },
  {"apply",       PC(ApplyCmd)     PF_PLAYING },
  {"lookat",      PC(LookAt)       PF_PLAYING },
  {"lock",        PC(LockItem)     PF_PLAYING },
  {"mark",        PC(MarkItem)     PF_PLAYING },
  {"move",        PC(MoveCmd)      PF_PLAYING },
  {"ext",         PC(ExtCmd)       0 }, // CF+/Deliantra
  {"mapredraw",   PC(MapRedrawCmd) 0 },
  {"mapinfo",     PC(MapInfoCmd)   0 }, // CF+/Deliantra

  {"reply",       SC(ReplyCmd)     0 },
  {"exti",        SC(ExtiCmd)      0 }, // CF+/Deliantra
  {"addme",       SC(AddMeCmd)     0 },
  {"askface",     SC(AskFaceCmd)   0 },
  {"requestinfo", SC(RequestInfo)  0 },
  {"setsound",    SC(SetSound)     0 },
  {"setup",       SC(SetUp)        0 },
  {"version",     SC(VersionCmd)   0 },

  {"toggleextendedtext", SC(NopCmd)   0 }, // deliantra: disabled
};

bool
client::may_execute (const packet_type *pkt) const
{
  return (!(pkt->flags & PF_PLAYER)  || pl)
      && (!(pkt->flags & PF_PLAYING) || state == ST_PLAYING);
}

// HACK: some commands currently should be executed
// even when the player is frozen. this hack detects
// those commands. it should be folded into may_execute,
// but kept seperate to emphasise the hack aspect, i.e.
// do it better, then remove.
static bool
always_immediate (const client *ns, const packet_type *pkt, const char *data, int len)
{
  if (!(pkt->flags & PF_COMMAND))
    return false;

  if (!ns->pl || !ns->pl->ob || !ns->pl->ob->map)
    return false;

  if (len >  4 && data [ 3] == ' ' && !strncmp (data, "say "              ,  4)) return true;
  if (len >  5 && data [ 4] == ' ' && !strncmp (data, "chat "             ,  5)) return true;
  if (len >  6 && data [ 5] == ' ' && !strncmp (data, "shout "            ,  6)) return true;
  if (len >  8 && data [ 7] == ' ' && !strncmp (data, "suicide "          ,  8)) return true;
  if (len > 18 && data [17] == ' ' && !strncmp (data, "accept-invitation ", 18)) return true;

  if (len == 7 && !strcmp (data, "suicide")) return true;

  return false;
}

void
client::execute (const packet_type *pkt, char *data, int datalen)
{
  if (may_execute (pkt) || always_immediate (this, pkt, data, datalen))
    {
      //TODO: only one format
      if (pkt->flags & PF_PLAYER)
        {
          ((void (*)(char *, int, player *))pkt->cb)((char *)data, datalen, pl);
          pl->need_updated_stats ();
        }
      else
        ((void (*)(char *, int, client *))pkt->cb)((char *)data, datalen, this);
    }
  else
    send_packet_printf ("msg %d log ERROR: you cannot execute '%s' now.", NDI_RED | NDI_REPLY, pkt->name);
}

bool
client::handle_packet ()
{
  int pkt_len = next_packet ();

  if (!pkt_len)
    return false;
  else if (pkt_len < 0)
    {
      LOG (llevError, "read error on player %s\n",
                      pl && pl->ob ? &pl->ob->name : "[anonymous]");
      destroy ();
      return false;
    }

  uint8_t save_byte = inbuf [pkt_len]; // rather ugly
  inbuf [pkt_len] = 0;      /* temporarily terminate buffer - useful for string data */

  /* First, break out beginning word.  There are at least
   * a few commands that do not have any paremeters.  If
   * we get such a command, don't worry about trying
   * to break it up.
   */
  int datalen;
  char *data = strchr ((char *)inbuf + 2, ' ');

  if (data)
    {
      *data++ = 0;
      datalen = pkt_len - (data - (char *)inbuf);
    }
  else
    {
      data = (char *)inbuf + 2; // better read garbage than segfault
      datalen = 0;
    }

  for (packet_type *pkt = packets; pkt < packets + ecb_array_length (packets); ++pkt)
    if (!strcmp ((char *)inbuf + 2, pkt->name))
      {
        if (pkt->flags & PF_PLAYER && !always_immediate (this, pkt, data, datalen))
          queue_command (pkt, data, datalen);
        else
          execute (pkt, data, datalen);

        goto next_packet;
      }

  // If we get here, we didn't find a valid command.
  send_packet_printf ("msg %d log ERROR: command '%s' not supported.", NDI_RED | NDI_REPLY, (char *)inbuf + 2);

next_packet:
  inbuf [pkt_len] = save_byte; // rather ugly
  skip_packet (pkt_len);

  // input buffer has space again
  socket_ev.poll (socket_ev.poll () | EV_READ);

  return true;
}

void
client::inbuf_handle ()
{
  if (!handle_packet ())
    return;

  while (handle_packet ())
    ;

  flush ();
}

// callback called when socket is either readable or writable
void
client::socket_cb (iow &w, int revents)
{
  //TODO remove when we have better socket cleanup logic
  if (destroyed ())
    {
      socket_ev.poll (0);
      return;
    }

  if (revents & EV_WRITE)
    {
      write_outputbuffer ();

      if (!outputbuffer.len)
        socket_ev.poll (socket_ev.poll () & ~EV_WRITE);
    }

  if (revents & EV_READ)
    {
      int amount;
      uint8 *rbuf;

      if (ws_version)
        {
          if (ws_inbuf_len + 2048 > ws_inbuf_alloc)
            ws_inbuf = (uint8 *)realloc (ws_inbuf, ws_inbuf_alloc += 4096);

          rbuf = ws_inbuf + ws_inbuf_len;
          amount = ws_inbuf_alloc - ws_inbuf_len;
        }
      else
        {
          rbuf = inbuf + inbuf_len;
          amount = sizeof (inbuf) - inbuf_len;
        }

      if (!amount)
        {
          // input buffer full
          LOG (llevError, "input buffer overflow.");
          destroy ();
          return;
        }

      amount = read (fd, rbuf, amount);

      if (!amount)
        {
          destroy ();
          return;
        }
      else if (amount < 0)
        {
          if (errno != EAGAIN && errno != EINTR)
            {
              LOG (llevError, "read error: %s\n", strerror (errno));
              destroy ();
              return;
            }

          // should not be here, normally
        }

      if (ws_version)
        {
          ws_inbuf_len += amount;

          while (ws_inbuf_len)
            {
              if (ws_inbuf_len < 2 + 4) // 6 is minimum length: op, len, mask
                break;

              int d   = 2;
              int fin = ws_inbuf [0] & 0x80;
              int op  = ws_inbuf [0] & 0x0f;
              int len = ws_inbuf [1] & 0x7f;

              if (len == 126)
                {
                  len = (ws_inbuf [2] << 8) | ws_inbuf [3];
                  d += 2;
                }
              else if (len == 127)
                {
                  if (ws_inbuf_len < 2 + 8)
                    return;

                  // we don't do extra long frames, if a browser wants to send >2**32 bytes,
                  // there are bigger issues to fix.
                  len = (ws_inbuf [8] << 8) |  ws_inbuf [9];
                  d += 8;
                }

              // we only continue if we have a complete frame
              if (ws_inbuf_len < d + 4 + len)
                return;

              switch (op)
                {
                  case  0: op = ws_inbuf_type; break; // continuation
                  case  1: ws_inbuf_type = 1;  break; // utf-8
                  case  2: ws_inbuf_type = 2;  break; // binary
                }

              if (len > sizeof (inbuf) - inbuf_len)
                {
                  // input buffer full
                  LOG (llevError, "input buffer overflow (ws).");
                  destroy ();
                  return;
                }

              uint8 *buf = inbuf + inbuf_len + 2;

              for (int i = 0; i < len; ++i)
                buf [i] = ws_inbuf [d + 4 + i] ^ ws_inbuf [d + (i & 3)];

              // remove frame
              ws_inbuf_len -= d + 4 + len;
              memmove (ws_inbuf, ws_inbuf + d + 4 + len, ws_inbuf_len);

              switch (op)
                {
                  case  1: // utf-8
                    // utf-8 encoded frames cannot have full length (MAXSOCKBUF) in all cases
                    // but we assume that these extra-long packets will be rare.
                    {
                      uint8 *a = buf;
                      uint8 *b = a;
                      uint8 *c = a + len;

                      for (; a < c; ++a, ++b)
                        {
                          *b = *a;

                          if (*a >= 0x80)
                            *b = (a [0] & 0x1f) << 6 | (a [1] & 0x3f), ++a;
                        }

                      len -= a - b;
                    }
                    /* FALLTHROUGH */
                  case  2: // binary
                    inbuf_len += len;

                    if (fin)
                      {
                        inbuf [0] = inbuf_len >> 8;
                        inbuf [1] = inbuf_len;

                        inbuf_len += 2;

                        inbuf_handle ();
                      }
                    break;

                  case 9: // ping
                    {
                      // send pong - we assume ping messages are <64k
                      // as we can't handle >10k at the moment anyway.
                      uint8 hdr [] = { 0x8a, 126, len >> 8, len };
                      send (hdr, sizeof (hdr));
                      send (inbuf + inbuf_len, len);
                    }
                    break;

                  case 10: // pong
                    break;

                  case 8: // close
                  default:
                    destroy ();
                    return;
                }
            }
        }
      else
        {
          inbuf_len += amount;
          inbuf_handle ();
        }
    }
}

/*******************************************************************************
 *
 * Start of write related routines.
 *
 ******************************************************************************/

/**
 * Adds data to a socket buffer for whatever reason.
 *
 * ns is the socket we are adding the data to, buf is the start of the
 * data, and len is the number of bytes to add.
 */
void
client::send (void *buf_, int len)
{
  char *buf = (char *)buf_;

  if (destroyed () || !buf)
    return;

  if (len + outputbuffer.len > SOCKETBUFSIZE)
    {
      LOG (llevDebug, "socket on fd %d has overrun internal buffer - marking as dead\n", fd);
      // shutdown the socket, this is safer than destroying it immediately
      // as lots of code in the callchain might still access the map etc.
      shutdown (fd, SHUT_RDWR);
      return;
    }

  int avail, end;

  /* data + end is where we start putting the new data.  The last byte
   * currently in use is actually data + end -1
   */
  end = outputbuffer.start + outputbuffer.len;
  /* The buffer is already in a wrapped state, so adjust end */
  if (end >= SOCKETBUFSIZE)
    end -= SOCKETBUFSIZE;

  avail = SOCKETBUFSIZE - end;

  /* We can all fit it behind the current data without wrapping */
  if (avail >= len)
    memcpy (outputbuffer.data + end, buf, len);
  else
    {
      memcpy (outputbuffer.data + end, buf, avail);
      memcpy (outputbuffer.data, buf + avail, len - avail);
    }

  outputbuffer.len += len;
}

/**
 * Takes a string of data, and writes it out to the socket. A very handy
 * shortcut function.
 */
void
client::send_packet (packet &sl)
{
  if (destroyed ())
    return;

  if (sl.length () > MAXSOCKBUF)
    {
      LOG (llevError, "Trying to send a buffer beyond properly size, len =%d\n", sl.length ());
      /* Almost certainly we've overflowed a buffer, so quit now to make
       * it easier to debug.
       */
      abort ();
    }

  if (!sl.length ())
    return;

  if (ws_version == 8)
    {
      static uint8 buf [MAXSOCKBUF * 2 + 4];

      uint8 *b = buf + 4;
      for (uint8 *a = sl.buf_ + sl.hdrlen; a < sl.cur; ++a)
        {
          if (*a < 0x80)
            *b++ = *a;
          else
            {
              *b++ = 0xc0 | ((*a >> 6) & 0x1f);
              *b++ = 0x80 | ( *a       & 0x3f);
            }
        }

      assert (b - buf < sizeof (buf));

      int len = b - (buf + 4);

      if (len < 126)
        {
          buf [2] = 0x81;
          buf [3] = len;

          send (buf + 2, len + 2);
        }
      else
        {
          buf [0] = 0x81;
          buf [1] = 126;
          buf [2] = len >> 8;
          buf [3] = len;

          send (buf, len + 4);
        }
    }
  else if (ws_version == 13)
    {
      int len = sl.length ();

      if (len < 126)
        {
          uint8 hdr [] = { 0x82, len };
          send (hdr, sizeof (hdr));
        }
      else
        {
          uint8 hdr [] = { 0x82, 126, len >> 8, len };
          send (hdr, sizeof (hdr));
        }

      send (sl.buf_ + sl.hdrlen, len);
    }
  else
    {
      assert (sl.hdrlen == 2);

      sl.buf_ [0] = sl.length () >> 8;
      sl.buf_ [1] = sl.length ()     ;

      send (sl.buf_, sl.length () + sl.hdrlen);
    }
}

void
client::send_packet (const char *buf, int len)
{
  packet sl;

  sl << data_n (buf, len);
  send_packet (sl);
}

void
client::send_packet (const char *buf)
{
  send_packet (buf, strlen (buf));
}

void
client::send_packet_printf (const char *format, ...)
{
  packet sl;

  va_list ap;
  va_start (ap, format);
  sl.vprintf (format, ap);
  va_end (ap);

  send_packet (sl);
}

void
client::send_msg (int color, const char *type, const char *msg)
{
  if (!msg || !type) // it can happen (for example, missing attack messages cause this)
    {
      LOG (logBacktrace | llevError, "send_msg(%d,%p,%p) called with NULL msg or type.\n", color, type, msg);
      return;
    }

  int len = strlen (msg);

  if (!(color & NDI_VERBATIM)
      && (msg_is_special (msg)
          || (type [0] == 'c' && type [1] == '/') || len > (MAXSOCKBUF - 128)))
    cfperl_send_msg (this, color, type, msg);
  else
    send_packet_printf ("msg %d %s %s", color & NDI_CLIENT_MASK, type, msg);
}

/***********************************************************************
 *
 * packet functions/utilities
 *
 **********************************************************************/

packet::packet (const char *name)
{
  reset ();

  int len = strlen (name);
  memcpy (cur, name, len); cur += len;
  *cur++ = ' ';
}

packet &packet::operator <<(const ber32 v)
{
  enum { maxlen = 32 / 7 + 1 };
  uint8 buf[maxlen];
  uint8 *p = buf + maxlen;
  uint32 val = v.val;

  *--p = val & 0x7F;

  while (val > 0x7F)
    {
      val >>= 7;
      *--p = (val & 0x7F) | 0x80;
    }

  return *this << data_n (p, buf + maxlen - p);
}

packet &packet::operator <<(const data_n &v)
{
  if (room () < v.len)
    reset ();
  else
    {
      if (v.len)
        {
          memcpy (cur, v.ptr, v.len);
          cur += v.len;
        }
    }

  return *this;
}

packet &packet::operator <<(const data8  &v)
{
  unsigned int len = min (v.len, 0x00FF);
  return *this << uint8  (len) << data_n (v.ptr, len);
}

packet &packet::operator <<(const data16 &v)
{
  unsigned int len = min (v.len, 0xFFFF);
  return *this << uint16 (len) << data_n (v.ptr, len);
}

packet &packet::operator <<(const char *v)
{
  return *this << data_n (v, v ? strlen (v) : 0);
}

void
packet::vprintf (const char *format, va_list ap)
{
  int size = room ();

  int len = vsnprintf ((char *)cur, size, format, ap);

  if (len >= size)
    return reset ();

  cur += len;
}

